// Options.java
// Copyright (C) 2008 by Jeff Gold.
//
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see
// <http://www.gnu.org/licenses/>.
//
// ---------------------------------------------------------------------
// There are a few things this class should do that it doesn't.  First,
// support for dynamic help strings in Base is currently missing.
// Second, Group doesn't support help strings or short options.  Third,
// the help output doesn't set groups apart with white space.  Fourth,
// there is no support for longest prefix match.
package net.esclat.ripple;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.TreeSet;
import java.util.LinkedList;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.BufferedReader;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

/** Generic option parsing with support for command line and
 *  confuration files. */
public class Options {

    /** Base exception for problems with option parsing. */
    public static class Problem extends RuntimeException {
        java.lang.String opt = null;
        java.lang.String arg = null;

        public Problem(Throwable cause) { super(cause); }
        public Problem(java.lang.String msg) { super(msg); }
        public Problem(java.lang.String msg,
                       java.lang.String opt)
        { super(msg + " " + opt); this.opt = opt; }
        public Problem(java.lang.String msg,
                       java.lang.String opt,
                       java.lang.String arg) {
            super(msg + " for " + opt + ": " + arg);
            this.opt = opt; this.arg = arg;
        }
    }

    /** Abstraction for one or more configurable parameters. */
    public static interface Option {
        /** Returns a list of names for which a help message is
         *  provided by this option. */
        public Iterable<java.lang.String> names();

        /** Returns a brief (sixty characters or less) help message
         *  for the given option name. */
        public java.lang.String help(java.lang.String opt);

        /** Returns the cannonical name for an option name or null if
         *  this object doesn't support that. */
        public java.lang.String find(java.lang.String opt);

        /** Returns the cannonical name for an option character or
         *  null if this object doesn't support that. */
        public java.lang.String find(char shr);
        
        /** Associate the argument with the named option.  When
         *  doing so is not possible a <code>Problem</code>
         *  exception or a subclass is expected to be thrown. */
        public void convert(java.lang.String opt,
                            java.lang.String arg);

        /** Returns true iff the option can be used without an
         *  argument. */
        public boolean standalone(java.lang.String opt);
    }

    private ArrayList<Option> opts = new ArrayList<Option>();
    public Options() {}
    public <T extends Option> T add(T s) { opts.add(s); return s; }

    protected static java.lang.String tagArg = "=";
    protected static java.lang.String tagOpt = "--";
    protected static java.lang.String tagShr = "-";

    private class NamedOption {
        public Option o;
        public java.lang.String name;

        public NamedOption(Option o, java.lang.String name)
        { this.o = o;  this.name = name; }
    }

    public Iterable<java.lang.String> parseArgs
        (Iterable<java.lang.String> args) {
        ArrayList<java.lang.String> extra =
            new ArrayList<java.lang.String>();
        LinkedList<NamedOption> wanters =
            new LinkedList<NamedOption>();
        boolean fill = false;
        for (java.lang.String arg : args) {
            if (fill) {
                extra.add(arg);
            } else if (wanters.size() > 0) {
                NamedOption next = wanters.remove();
                next.o.convert(next.name, arg);
            } else if (arg.equals(tagOpt)) {
                fill = true;
            } else if (arg.startsWith(tagOpt)) {
                boolean matched = false;
                for (Option o : opts) {
                    java.lang.String opt = null;
                    matched = true;
                    int marg = arg.indexOf(tagArg);
                    if ((marg >= 0) &&
                        ((opt = o.find(arg.substring(tagOpt.length(),
                                                     marg))) != null)) {
                        o.standalone(opt);
                        o.convert(opt, arg.substring(marg + 1));
                    } else if (null != (opt = o.find
                                        (arg.substring
                                         (tagOpt.length())))) {
                        if (!o.standalone(opt))
                            wanters.add(new NamedOption(o, opt));
                    } else matched = false;
                    if (matched) break;
                }
                if (!matched)
                    throw new Problem("Unknown option", arg);
            } else if (arg.startsWith(tagShr)) {
                for (char c : arg.substring
                         (tagShr.length()).toCharArray()) {
                    boolean matched = false;
                    for (Option o : opts) {
                        java.lang.String opt;
                        if ((opt = o.find(c)) != null) {
                            if (!o.standalone(opt))
                                wanters.add(new NamedOption(o, opt));
                            matched = true;
                            break;
                        }
                    }
                    if (!matched)
                        throw new Problem("Unknown option",
                                          new Character(c).toString());
                }
            } else extra.add(arg);
        }
        if (wanters.size() > 0)
            throw new Problem("Missing argument", wanters.peek().name);
        return extra;
    }

    public Iterable<java.lang.String> parseArgs
        (java.lang.String[] args)
    { return parseArgs(Arrays.asList(args)); }

    public void parseConfig(BufferedReader reader) throws IOException {
        java.lang.String line;
        while ((line = reader.readLine()) != null) {
            if (line.matches("^\\s*(#.*)?")) // empty and comments lines
                continue;
            java.lang.String opt;
            boolean matched = false;
            int point = line.indexOf(tagArg);
            if (point >= 0) {
                opt = line.substring(0, point).trim();
                java.lang.String arg =
                    line.substring(point + tagArg.length()).trim();
                if ((arg.length() > 1) &&
                    ((arg.startsWith("\"") && arg.endsWith("\"")) ||
                     (arg.startsWith("'") && arg.endsWith("'"))))
                    arg = arg.substring(1, arg.length() - 2);
                for (Option o : opts) {
                    if (o.find(opt) != null) {
                        o.standalone(opt);
                        o.convert(opt, arg);
                        matched = true;
                        break;
                    }
                }
            } else {
                opt = line.trim();
                for (Option o : opts) {
                    if (o.find(opt) != null) {
                        if (!o.standalone(opt))
                            throw new Problem
                                ("Missing argument for", opt);
                        matched = true;
                        break;
                    }
                }
            }
            if (!matched)
                throw new Problem("Unknown setting: ", opt);
        }
    }

    public void parseConfig(InputStreamReader stream) throws IOException
    { parseConfig(new BufferedReader(stream)); }

    public void parseConfig(InputStream stream) throws IOException
    { parseConfig(new BufferedReader(new InputStreamReader(stream))); }

    public void help(java.io.PrintStream stream) {
        int maxlen = 0;
        for (Option o : opts)
            for (java.lang.String opt : o.names())
                maxlen = Math.max(maxlen, opt.length());
        for (Option o : opts)
            for (java.lang.String opt : o.names())
                stream.printf("%s - %s", opt, o.help(opt)).println();
        // FIXME: automatically wrap help strings
    }

    /** Common resources for supporting many familiar option types. */
    public static class Base implements Option {
        protected java.lang.String opt;
        protected char             shr;
        protected java.lang.String hlp;
        public boolean applied = false;
        
        /** Creates a degenerate Option suitable for subclass use. */
        protected Base() { this(null, '\0', null); }
        /** Creates an option. */
        public Base(java.lang.String opt, char shr,
                    java.lang.String help)
        { this.opt = opt;  this.shr = shr; this.hlp = help; }
        /** Creates an option with no short form. */
        public Base(java.lang.String opt,
                    java.lang.String help) { this(opt, '\0', help); }
        /** Creates an option with no help. */
        public Base(java.lang.String opt, char shr)
        { this(opt, shr, null); }
        /** Creates an option with no short form and no help. */
        public Base(java.lang.String opt) { this(opt, '\0', null); }

        /** Implements Option. */
        public java.lang.String find(java.lang.String opt) {
            return ((this.opt != null) && (this.opt.equals(opt))) ?
                opt : null;
        }

        /** Implements Option. */
        public java.lang.String find(char shr) {
            return ((this.shr != '\0') && (this.shr == shr)) ?
                opt : null;
        }

        /** Implements Option. */
        public Iterable<java.lang.String> names() {
            ArrayList<java.lang.String> result =
                new ArrayList<java.lang.String>();
            if (hlp != null) result.add(this.opt);
            return result;
        }

        /** Implements Option. */
        public java.lang.String help(java.lang.String opt)
        { return hlp; }

        /** Implements Option. */
        public void convert(java.lang.String opt,
                            java.lang.String arg)
        { throw new Problem("Unnecessary argument", opt, arg); }

        /** Implements Option. */
        public boolean standalone(java.lang.String opt)
        { return (applied = true); }
    }
    
    /** Extensible groups of options using reflection. */
    public static class Group implements Option {
        /** Creates a degenerate Group suitable for subclass use. */
        protected Group() {}

        /** Returns an option specific standalone method or null.. */
        private Method getStandalone(java.lang.String opt) {
            Method result = null;
            try { return getClass().getMethod("standalone_" + opt); }
            catch (NoSuchMethodException ex) {}
            return result;
        }
        /** Returns an option specific convert method or null. */
        private Method getConvert(java.lang.String opt) {
            Method result = null;
            try { return getClass().getMethod("convert_" + opt,
                                              java.lang.String.class); }
            catch (NoSuchMethodException ex) {}
            return result;
        }

        /** Implements Option. */
        public java.lang.String find(java.lang.String opt) {
            return ((getConvert(opt) != null) ||
                    (getStandalone(opt) != null)) ? opt : null;
        }

        /** Implements Option. */
        public java.lang.String find(char shr) {
            // FIXME: support sub-options with short names too.
            return null;
        }

        /** Implements Option. */
        public Iterable<java.lang.String> names() {
            ArrayList<java.lang.String> result =
                new ArrayList<java.lang.String>();
            // FIXME: find dynamic help strings.
            return result;
        }

        /** Implements Option. */
        public java.lang.String help(java.lang.String opt) {
            // FIXME: find dynamic help strings.
            return null;
        }

        /** Implements Option. */
        public void convert(java.lang.String opt,
                            java.lang.String arg) {
            Method m = getConvert(opt);
            if (m != null)
                try {
                    m.invoke(this, arg);
                    return;
                } catch (IllegalAccessException ex) {
                } catch (InvocationTargetException ex) {
                    Throwable cause = ex.getCause();
                    if (cause instanceof Problem)
                        throw (Problem)cause;
                    else throw new Problem(cause);
                }
            throw new Problem("Unnecessary argument", opt, arg);
        }

        /** Implements Option. */
        public boolean standalone(java.lang.String opt) {
            Method m = getStandalone(opt);
            if (m != null)
                try {
                    m.invoke(this);
                    return true;
                } catch (IllegalAccessException ex) {
                } catch (InvocationTargetException ex) {
                    Throwable cause = ex.getCause();
                    if (cause instanceof Problem)
                        throw (Problem)cause;
                    else throw new Problem(cause);
                }
            return getConvert(opt) == null;
        }
    }
    
    /** A simple switch setting.  A public <code>value</code>
     *  field permits access to the parsed result. */
    public static class Switch extends Base {
        public boolean value = false;
        public Switch(java.lang.String opt) { super(opt); }
        public boolean standalone(java.lang.String opt)
        { value = true; return true; }
    }

    /** A simple counter setting.  A public <code>value</code>
     *  field permits access to the parsed result. */
    public static class Counter extends Base {
        public int value;
        public Counter(java.lang.String opt, int defl)
        { super(opt);  value = defl; }
        public Counter(java.lang.String opt) { this(opt, 0); }
        public boolean standalone(java.lang.String opt)
        { value++; return true; }
    }

    /** A simple string setting.  A public <code>value</code>
     *  field permits access to the parsed result. */
    public static class String extends Base {
        public java.lang.String value;
        public String(java.lang.String opt, java.lang.String defl)
        { super(opt);  value = defl; }
        public String(java.lang.String opt) { this(opt, null); }
        public boolean standalone(java.lang.String opt)
        { return false; }
        public void convert(java.lang.String opt,
                            java.lang.String arg) { value = arg; }
    }

    /** A simple integer setting.  A public <code>value</code>
     *  field permits access to the parsed result. */
    public static class Integer extends Base {
        public int value;
        public Integer(java.lang.String opt, int defl)
        { super(opt);  value = defl; }
        public Integer(java.lang.String opt) { this(opt, 0); }
        public boolean standalone(java.lang.String opt)
        { return false; }
        public void convert(java.lang.String opt,
                            java.lang.String arg)
        { value = check(opt, arg); }

        public static int check(java.lang.String opt,
                            java.lang.String arg) {
            try { return java.lang.Integer.parseInt(arg); }
            catch (NumberFormatException ex) {
                throw new Problem("Argument must be an integer",
                                  opt, arg);
            }
        }
    }

    /** A simple integer setting.  A public <code>value</code>
     *  field permits access to the parsed result. */
    public static class Long extends Base {
        public long value;
        public Long(java.lang.String opt, long defl)
        { super(opt);  value = defl; }
        public Long(java.lang.String opt) { this(opt, 0); }
        public boolean standalone(java.lang.String opt)
        { return false; }
        public void convert(java.lang.String opt,
                            java.lang.String arg)
        { value = check(opt, arg); }

        public static long check(java.lang.String opt,
                                 java.lang.String arg) {
            try { return java.lang.Long.parseLong(arg); }
            catch (NumberFormatException ex) {
                throw new Problem("Argument must be an integers",
                                  opt, arg);
            }
        }
    }

    /** An internet address setting.  A public <code>value</code>
     *  field permits access to the parsed result. */
    public static class Address extends Base {
        public InetAddress value;
        public Address(java.lang.String opt, InetAddress defl)
        { super(opt);  value = defl; }
        public Address(java.lang.String opt) { this(opt, null); }
        public boolean standalone(java.lang.String opt)
        { return false; }
        public void convert(java.lang.String opt,
                            java.lang.String arg)
        { value = check(opt, arg); }

        public static InetAddress check(java.lang.String opt,
                                        java.lang.String arg) {
            try {
                return InetAddress.getByName(arg);
            } catch (UnknownHostException ex) {
                throw new Problem("Host unknown", opt, arg);
            }
        }
    }    

    /** A print stream setting.  A public <code>value</code> field
     *  permits access to the parsed result. */
    public static class PrintStream extends Base {
        public java.io.PrintStream value;
        public PrintStream(java.lang.String opt,
                           java.io.PrintStream defl)
        { super(opt);  value = defl; }
        public PrintStream(java.lang.String opt) { this(opt, null); }
        public boolean standalone(java.lang.String opt)
        { return false; }
        public void convert(java.lang.String opt,
                            java.lang.String arg)
        { value = check(opt, arg); }

        public static java.io.PrintStream check(java.lang.String opt,
                                                java.lang.String arg) {
            try {
                return new java.io.PrintStream
                    (new FileOutputStream(arg, true));
            } catch (FileNotFoundException ex) {
                throw new Problem("File not found", opt, arg);
            }
        }
    }    

    /** Configuration file setting.  The entire file is read and
     *  parsed at configuration time, so the order in which options
     *  are presented matters. */
    public static class Config extends Base {
        private Options opts;

        public Config(Options opts, java.lang.String opt, char shr,
                      java.lang.String hlp) {
            super(opt, shr, hlp);
            this.opts = opts;
        }
        public Config(Options opts, java.lang.String opt, char shr)
        { this(opts, opt, shr, null); }

        public void convert(java.lang.String opt,
                            java.lang.String arg) { doConfig(arg); }
        public boolean standalone(java.lang.String opt)
        { return false; }

        private void doConfig(java.lang.String arg) {
            try {
                opts.parseConfig(new FileInputStream(arg));
            } catch (FileNotFoundException ex) {
                throw new Problem("Configuration", arg,
                                  ex.getMessage());
            } catch (IOException ex) {
                throw new Problem("Configuration", arg,
                                  ex.getMessage());
            } catch (Problem ex) {
                throw new Problem("Configuration", arg,
                                  ex.getMessage());
            }
        }
    }

    /** A simple unit test. */
    public static void main(java.lang.String[] args) {
        Options opts   = new Options();
        Integer apple  = opts.add(new Integer("apple"));
        String  banana = opts.add(new String("banana"));

        class Color extends Group {
            public int shiny = 0;
            public int red   = 0;
            public int green = 0;
            public int blue  = 0;

            public void standalone_shiny() { shiny++; }
            public void convert_red(java.lang.String arg)
            { red = Integer.check("red", arg); }
            public void convert_green(java.lang.String arg)
            { green = Integer.check("green", arg); }
            public void convert_blue(java.lang.String arg)
            { blue = Integer.check("blue", arg); }
        }
        Color c = opts.add(new Color());

        opts.parseArgs(args);
        System.out.println("apple  = " + apple.value);
        System.out.println("banana = " + banana.value);
        System.out.println("shiny = " + c.shiny);
        System.out.println("red = " + c.red);
        System.out.println("green = " + c.green);
        System.out.println("blue = " + c.blue);
    }
}
